#version 130
#define USE_HIGH_QUALITY_BLUR
#define ENABLE_FSAA
#define ENABLE_SSAO
#define ENABLE_BLOOM
#define DEPTH_OF_FIELD
// #define UNDERWATER

uniform sampler2D texture;
uniform sampler2D depthtexture;
uniform ivec3 submerged;
uniform vec3 focal;
uniform vec3 frameRadian;
uniform vec3 screendata;
in vec4 texCoord;
in vec4 depthCoord;

// HYPERFOCAL = (Focal Distance ^ 2)/(Circle of Confusion * F Stop) + Focal Distance
const float HYPERFOCAL   = 4.0; //3.132;
const float PICONSTANT = 3.14159265357989;
const float BLOOM_OFFSET   = 4.0;
const float BLOOM_STRENGTH = 0.25;
const float PI2                 = 6.283185307179586476925286766559;
const float EXPOSURE = 0.15;
const float UNDERWATER_CAP = 0.9;

float getDepth(vec2 coord);
float getCursorDepth(vec2 coord);
float expDepth2linearDepth(vec2 coord);
vec4 getBlurredColor();
vec4 getSample(vec2 coord, vec2 aspectCorrection);
vec4 getSampleWithBoundsCheck(vec2 offset);
vec4 getAA();
vec4 getBloom(vec4 color);
vec4 getSSAO();

float samples = 0.0;
float far;
const float near =   0.1;
vec2 space;

const float FOG_DENSITY   =  2.0;
const float FOG_RAMP      =  0.9;
const float FOG_LUMINANCE =  0.5;
const float FOG_MAXIMUM   =  1.0;
const float FOG_MINIMUM   =  0.25;
const float FOG_DIST      = 32.0;

vec4 hjemmebakt_fog(vec4 color, float depth) {
	// submersed fog
	float f = min(FOG_MAXIMUM, FOG_MINIMUM + pow(min(1.0, depth / FOG_DIST), FOG_RAMP) * FOG_DENSITY);
	float lum = min(1.0, (0.5 + 0.2126 * color.r + 0.7152 * color.g + 0.0722 * color.b) * FOG_LUMINANCE);
	return vec4(mix(color.rgb, vec3(0.0, 0.0, lum), f), 1.0);
}

void main() {

	far = focal[2];
	vec2  coord = texCoord.st;
	float depth = getDepth(coord);

	#ifdef ENABLE_FSAA
		vec4 baseColor = getAA();	
	#else
		vec4 baseColor = texture2D(texture, coord);
	#endif

	float mixAmount = 0.0;

#ifdef DEPTH_OF_FIELD
	//HYPERFOCAL = getCursorDepth(vec2(0.5, 0.5)) * 3.132; //real hyperfocal
	float cursorDepth = focal[0]; //getCursorDepth(vec2(0.5));

	// foreground blur = 1/2 background blur. Blur should follow exponential pattern until cursor = hyperfocal -- Cursor before hyperfocal
	// Blur should go from 0 to 1/2 hyperfocal then clear to infinity -- Cursor @ hyperfocal.
	// hyperfocal to inifity is clear though dof extends from 1/2 hyper to hyper -- Cursor beyond hyperfocal

	if (depth < cursorDepth) {
			mixAmount = clamp(2.0 * ((clamp(cursorDepth, 0.0, HYPERFOCAL) - depth) / (clamp(cursorDepth, 0.0, HYPERFOCAL))), 0.0, 1.0);
} else if (cursorDepth == HYPERFOCAL) {
			mixAmount = 0.0;
} else {
			mixAmount = 1.0 - clamp((((cursorDepth * HYPERFOCAL) / (HYPERFOCAL - cursorDepth)) - (depth - cursorDepth)) / ((cursorDepth * HYPERFOCAL) / (HYPERFOCAL - cursorDepth)), 0.0, 1.0);
}
#endif

#ifdef ENABLE_SSAO
	baseColor *= getSSAO();
#endif

	const float BLURSTRENGTH = 1.4;
	float BLURDIST = focal[2] * 0.5;

	if (depth > BLURDIST) mixAmount += pow( (depth - BLURDIST) / (focal[2] - BLURDIST), 0.75) * BLURSTRENGTH;

	// if (mixAmount != 0.0) {
	baseColor = mix(baseColor, getBlurredColor(), clamp(mixAmount, 0.25, 0.85));
	// }

#ifdef ENABLE_BLOOM
	baseColor = getBloom(baseColor);
#endif

	// apply exposure
	// baseColor = vec4( vec3(1.0 - exp2(-baseColor.rgb * EXPOSURE)), 1.0);
	// output color
	if (submerged[1] == 1) {
		gl_FragColor = hjemmebakt_fog(baseColor, depth);
	} else {
		gl_FragColor = baseColor;
	}
}


float getDepth(vec2 coord) {
    return 2.0 * near * far / (far + near - (2.0 * texture2D(depthtexture, coord).x - 1.0) * (far - near));
}

float expDepth2linearDepth(vec2 coord) {
   return ( 2.0 * near ) / ( far + near - texture2D(depthtexture, coord).x * ( far - near ) );
}

float getCursorDepth(vec2 coord) {
	return 2.0 * near * far / (far + near - (2.0 * texture2D(depthtexture, coord).x - 1.0) * (far - near));
}

vec4 getBlurredColor() {
	vec4 blurredColor = vec4(0.0);
	float depth = getDepth(depthCoord.xy);
	vec2 aspectCorrection = vec2(1.0 * 0.005, screendata[2] * 0.005);

	vec2 ac0_4 = 0.4 * aspectCorrection;	// 0.4
#ifdef USE_HIGH_QUALITY_BLUR
	vec2 ac0_4x0_4 = 0.4 * ac0_4;			// 0.16
	vec2 ac0_4x0_7 = 0.7 * ac0_4;			// 0.28
#endif
	
	vec2 ac0_29 = 0.29 * aspectCorrection;	// 0.29
#ifdef USE_HIGH_QUALITY_BLUR
	vec2 ac0_29x0_7 = 0.7 * ac0_29;			// 0.203
	vec2 ac0_29x0_4 = 0.4 * ac0_29;			// 0.116
#endif

	vec2 ac0_15 = 0.15 * aspectCorrection;	// 0.15
	vec2 ac0_37 = 0.37 * aspectCorrection;	// 0.37
#ifdef USE_HIGH_QUALITY_BLUR
	vec2 ac0_15x0_9 = 0.9 * ac0_15;			// 0.135
	vec2 ac0_37x0_9 = 0.37 * ac0_37;		// 0.1369
#endif

	vec2 lowSpace = texCoord.st;
	vec2 highSpace = 1.0 - lowSpace;
	space = vec2(min(lowSpace.s, highSpace.s), min(lowSpace.t, highSpace.t));

	if (space.s >= ac0_4.s && space.t >= ac0_4.t) {

		blurredColor += texture2D(texture, texCoord.st + vec2(0.0, ac0_4.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_4.s, 0.0));   
		blurredColor += texture2D(texture, texCoord.st + vec2(0.0, -ac0_4.t)); 
		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_4.s, 0.0)); 
		
#ifdef USE_HIGH_QUALITY_BLUR
		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_4x0_7.s, 0.0));       
		blurredColor += texture2D(texture, texCoord.st + vec2(0.0, -ac0_4x0_7.t));     
		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_4x0_7.s, 0.0));     
		blurredColor += texture2D(texture, texCoord.st + vec2(0.0, ac0_4x0_7.t));

		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_4x0_4.s, 0.0));
		blurredColor += texture2D(texture, texCoord.st + vec2(0.0, -ac0_4x0_4.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_4x0_4.s, 0.0));
		blurredColor += texture2D(texture, texCoord.st + vec2(0.0, ac0_4x0_4.t));
#endif

		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_29.s, -ac0_29.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_29.s, ac0_29.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_29.s, ac0_29.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_29.s, -ac0_29.t));
	
#ifdef USE_HIGH_QUALITY_BLUR
		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_29x0_7.s, ac0_29x0_7.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_29x0_7.s, -ac0_29x0_7.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_29x0_7.s, ac0_29x0_7.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_29x0_7.s, -ac0_29x0_7.t));
		
		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_29x0_4.s, ac0_29x0_4.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_29x0_4.s, -ac0_29x0_4.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_29x0_4.s, ac0_29x0_4.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_29x0_4.s, -ac0_29x0_4.t));
#endif		
		
		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_15.s, ac0_37.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_37.s, ac0_15.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_37.s, -ac0_15.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_15.s, -ac0_37.t));

		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_15.s, ac0_37.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_37.s, ac0_15.t)); 
		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_37.s, -ac0_15.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_15.s, -ac0_37.t));

#ifdef USE_HIGH_QUALITY_BLUR
		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_15x0_9.s, ac0_37x0_9.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_37x0_9.s, ac0_15x0_9.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_37x0_9.s, -ac0_15x0_9.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_15x0_9.s, -ac0_37x0_9.t));

		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_15x0_9.s, ac0_37x0_9.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_37x0_9.s, ac0_15x0_9.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(-ac0_37x0_9.s, -ac0_15x0_9.t));
		blurredColor += texture2D(texture, texCoord.st + vec2(ac0_15x0_9.s, -ac0_37x0_9.t));
#endif

#ifdef USE_HIGH_QUALITY_BLUR
	    blurredColor *= 0.025;
#else
	    blurredColor *= 0.0625;
#endif
   
	} else {

		blurredColor += getSampleWithBoundsCheck(vec2(0.0, ac0_4.t));
		blurredColor += getSampleWithBoundsCheck(vec2(ac0_4.s, 0.0));   
		blurredColor += getSampleWithBoundsCheck(vec2(0.0, -ac0_4.t)); 
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_4.s, 0.0)); 

#ifdef USE_HIGH_QUALITY_BLUR
		blurredColor += getSampleWithBoundsCheck(vec2(ac0_4x0_7.s, 0.0));       
		blurredColor += getSampleWithBoundsCheck(vec2(0.0, -ac0_4x0_7.t));     
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_4x0_7.s, 0.0));     
		blurredColor += getSampleWithBoundsCheck(vec2(0.0, ac0_4x0_7.t));

		blurredColor += getSampleWithBoundsCheck(vec2(ac0_4x0_4.s, 0.0));
		blurredColor += getSampleWithBoundsCheck(vec2(0.0, -ac0_4x0_4.t));
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_4x0_4.s, 0.0));
		blurredColor += getSampleWithBoundsCheck(vec2(0.0, ac0_4x0_4.t));
#endif

		blurredColor += getSampleWithBoundsCheck(vec2(ac0_29.s, -ac0_29.t));
		blurredColor += getSampleWithBoundsCheck(vec2(ac0_29.s, ac0_29.t));
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_29.s, ac0_29.t));
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_29.s, -ac0_29.t));

#ifdef USE_HIGH_QUALITY_BLUR
		blurredColor += getSampleWithBoundsCheck(vec2(ac0_29x0_7.s, ac0_29x0_7.t));
		blurredColor += getSampleWithBoundsCheck(vec2(ac0_29x0_7.s, -ac0_29x0_7.t));
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_29x0_7.s, ac0_29x0_7.t));
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_29x0_7.s, -ac0_29x0_7.t));

		blurredColor += getSampleWithBoundsCheck(vec2(ac0_29x0_4.s, ac0_29x0_4.t));
		blurredColor += getSampleWithBoundsCheck(vec2(ac0_29x0_4.s, -ac0_29x0_4.t));
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_29x0_4.s, ac0_29x0_4.t));
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_29x0_4.s, -ac0_29x0_4.t));
#endif

		blurredColor += getSampleWithBoundsCheck(vec2(ac0_15.s, ac0_37.t));
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_37.s, ac0_15.t));
		blurredColor += getSampleWithBoundsCheck(vec2(ac0_37.s, -ac0_15.t));
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_15.s, -ac0_37.t));
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_15.s, ac0_37.t));
		blurredColor += getSampleWithBoundsCheck(vec2(ac0_37.s, ac0_15.t)); 
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_37.s, -ac0_15.t));
		blurredColor += getSampleWithBoundsCheck(vec2(ac0_15.s, -ac0_37.t));

#ifdef USE_HIGH_QUALITY_BLUR
		blurredColor += getSampleWithBoundsCheck(vec2(ac0_15x0_9.s, ac0_37x0_9.t));
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_37x0_9.s, ac0_15x0_9.t));
		blurredColor += getSampleWithBoundsCheck(vec2(ac0_37x0_9.s, -ac0_15x0_9.t));
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_15x0_9.s, -ac0_37x0_9.t));
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_15x0_9.s, ac0_37x0_9.t));
		blurredColor += getSampleWithBoundsCheck(vec2(ac0_37x0_9.s, ac0_15x0_9.t));
		blurredColor += getSampleWithBoundsCheck(vec2(-ac0_37x0_9.s, -ac0_15x0_9.t));
		blurredColor += getSampleWithBoundsCheck(vec2(ac0_15x0_9.s, -ac0_37x0_9.t));
#endif

	    blurredColor /= vec4(samples);

	}

    return blurredColor;
}

vec4 getSampleWithBoundsCheck(vec2 offset) {
	vec2 coord = texCoord.st + offset;
	if (coord.s <= 1.0 && coord.s >= 0.0 && coord.t <= 1.0 && coord.t >= 0.0) {
		samples += 1.0;
		return texture2D(texture, coord);
	} else {
		return vec4(0.0);
	}
}

#ifdef ENABLE_SSAO
vec4 getSSAO() {

const float SSAO_CAP            =  0.5;
const float SSAO_MULTIPLY       =  4.0;
const float SSAO_DEPTHTOLERANCE =  0.05;
const int   SSAO_SAMPLES        =  2;

	float depth = expDepth2linearDepth(texCoord.st);

	float d = 0.0;
	float spw = 2.0 / screendata[0];
	float sph = 2.0 / screendata[1];
	float pw = spw;
	float ph = sph;

	float ao = 0.0, totalao = 0.0;
	float aoMultiplier  = SSAO_MULTIPLY;
	float aoTolerance   = SSAO_DEPTHTOLERANCE * (0.25 + 0.75 * depth);

	int j = 0;
	for (int i; i < SSAO_SAMPLES; i++) {

	   d = depth - expDepth2linearDepth( vec2(texCoord.s+pw,texCoord.t));  // +x
	   ao+= max(0.0, d-aoTolerance) * aoMultiplier;

	   d = depth - expDepth2linearDepth( vec2(texCoord.s-pw,texCoord.t));  // -x
	   ao+= max(0.0, d-aoTolerance) * aoMultiplier;

	   d = depth - expDepth2linearDepth( vec2(texCoord.s,texCoord.t+ph));  // +y
	   ao+= max(0.0, d-aoTolerance) * aoMultiplier;

	   d = depth - expDepth2linearDepth( vec2(texCoord.s,texCoord.t-ph));  // -y
	   ao+= max(0.0, d-aoTolerance) * aoMultiplier;

	   d = depth - expDepth2linearDepth( vec2(texCoord.s+pw,texCoord.t+ph));  //+x +y
	   ao+= max(0.0, d-aoTolerance) * aoMultiplier;

	   d = depth - expDepth2linearDepth( vec2(texCoord.s+pw,texCoord.t-ph));  //+x -y
	   ao+= max(0.0, d-aoTolerance) * aoMultiplier;

	   d = depth - expDepth2linearDepth( vec2(texCoord.s-pw,texCoord.t-ph)); // -x -y
	   ao+= max(0.0, d-aoTolerance) * aoMultiplier;

	   d = depth - expDepth2linearDepth( vec2(texCoord.s-pw,texCoord.t+ph)); // -x +y
	   ao+= max(0.0, d-aoTolerance) * aoMultiplier;

	pw *= 2.0;
	ph *= 2.0;
	aoMultiplier /= 4.0;
	totalao += ao / 8.0;

	}

	totalao /= SSAO_SAMPLES * 2.0;
	//if (totalao > 0.1) {
		return vec4(1.0 - clamp(totalao * totalao, 0.0, SSAO_CAP));
	//} else {
	//	return vec4(1.0);
	//}
}
#endif

#ifdef ENABLE_BLOOM

vec4 getBloom(vec4 color) {
	vec4 sum = vec4(0.0);
	vec2 texcoord = texCoord.xy;
	
	const float BLOOM_POWER = 0.25;
	const float BLOOM_SIZE  = 1.5;
	vec2 offset = vec2(BLOOM_SIZE / screendata[0], BLOOM_SIZE / screendata[1]);
	
	int i, j;
	for(i = -2; i < 3; i++) {
		for (j = -2; j < 3; j++) {
			sum += texture2D(texture, texcoord + vec2(i, j) * offset) * BLOOM_POWER;
		}
	}
	float strength = length(sum);
	if (strength > 4.0)
		strength = 1.0;
	else
		strength /= 4.0;
	return sum*sum*sum*0.0080 * strength + color;
}

vec4 getBloomB(vec4 color) {
	
	vec4 sum = vec4(0.0);
	vec2 texcoord = texCoord.xy;
	const float BLOOM_SIZE = 4.0;
	vec2 offset   = vec2(BLOOM_SIZE / screendata[0], BLOOM_SIZE / screendata[1]);
	
	for(int i = -3; i < 4; i++) { // Lower loop count for performance
		for (int j = -2; j < 3; j++) {
			sum += texture2D( texture, texcoord + vec2( j, i ) * offset ) * 0.25;
		}
	}
	// sum *= 0.1;
	
	if (color.r < 0.3) {
		sum = sum * sum * 0.012 + color;
	} else {
		if (color.r < 0.5) {
			sum = sum * sum * 0.009 + color;
		} else {
			sum = sum * sum * 0.0075 + color;
		}
	}
	return sum;
}
#endif

#ifdef ENABLE_FSAA
#define KERNEL_SIZE 9

// Gaussian kernel
// 1 2 1
// 2 4 2
// 1 2 1	
float kernel[KERNEL_SIZE];
vec2  offset[KERNEL_SIZE];

vec4 getAA() {

	float depth = expDepth2linearDepth(texCoord.xy);
	float step_w = 0.5 / screendata[0];
	float step_h = 0.5 / screendata[1];
	float depthThreshold = 0.1 * pow(depth, 1.5);

	if ( abs( depth - expDepth2linearDepth( texCoord.xy + vec2( step_w, 0 ) ) ) > depthThreshold || 
		 abs( depth - expDepth2linearDepth( texCoord.xy + vec2( -step_w, 0 ) ) ) > depthThreshold || 
		 abs( depth - expDepth2linearDepth( texCoord.xy + vec2( 0, step_h ) ) ) > depthThreshold || 
		 abs( depth - expDepth2linearDepth( texCoord.xy + vec2( 0, -step_h ) ) ) > depthThreshold ) {
		
		kernel[0] = 1.5/16.0;		kernel[1] = 2.0/16.0;		kernel[2] = 1.5/16.0;
		kernel[3] = 2.0/16.0;		kernel[4] = 2.0/16.0;		kernel[5] = 2.0/16.0;
		kernel[6] = 1.5/16.0;		kernel[7] = 2.0/16.0;		kernel[8] = 1.5/16.0;

		offset[0] = vec2(-step_w, -step_h);
		offset[1] = vec2(0.0, -step_h);
		offset[2] = vec2(step_w, -step_h);

		offset[3] = vec2(-step_w, 0.0);
		offset[4] = vec2(0.0, 0.0);
		offset[5] = vec2(step_w, 0.0);
		
		offset[6] = vec2(-step_w, step_h);
		offset[7] = vec2(0.0, step_h);
		offset[8] = vec2(step_w, step_h);
		
		vec4 sum = vec4(0.0);
		for(int i=0; i<KERNEL_SIZE; i++ ) {
			sum += texture2D(texture, texCoord.st + offset[i]) * kernel[i];
		}

		return sum; // vec4(1.0, 1.0, 1.0, 0.0);
   } else {
     return texture2D(texture, texCoord.st);
   }
}
#endif
